#include <owl.h>
#include <windowsx.h>

#include "ids.h"
#include "showdraw.h"
#include "gridview.h"
#include "hwxbrowz.h"
#include "drawdata.h"


//
//  Constructor of Grid-Window
//
TGridWindow::TGridWindow( PTWindowsObject AParent, int AnId, PTDrawData pd, 
                          PTModule AModule ) :
TControl( AParent, AnId, AModule )
{
    pDrawData    = pd;

    vMode        = gridMode;

    wVisibleCols = wVisibleRows =  5;
    wLeftCol     = wTopRow      =  0;

    wSelectedCell   = 0;
    hSolidPen       = CreatePen( PS_SOLID, 1, RGB( 0x80, 0x80, 0x80 ));

    SetNumRows( 10 );
    SetNumCols( wVisibleCols );
}



//
//  Destructor - performs cleanup
//
TGridWindow::~TGridWindow()
{
    DeletePen( hSolidPen );
}



//
//  Initialization routine: called just prior to
//  displaying the GridWindow.
//
void TGridWindow::SetupWindow( void )
{
    TControl::SetupWindow();

    RECT rect;
    WORD wWidth, wHeight;

    GetClientRect( HWindow, &rect );
    wWidth = rect.right - rect.left;
    wHeight= rect.bottom- rect.top;

    wCellWidth = wWidth /wVisibleCols;
    wCellHeight= wHeight/wVisibleRows;

    wBdrWidth  = ( wWidth % wVisibleCols )/2;
    wBdrHeight = ( wHeight% wVisibleRows )/2;
}



//
//  Allows one to specify particular class styles
//
void TGridWindow::GetWindowClass( WNDCLASS & AWndClass )
{
    TControl::GetWindowClass( AWndClass );
    AWndClass.style |= CS_DBLCLKS;
}




//
//  Routine to update the cells which viewing characters
//  in the Grid format...
//
void TGridWindow::UpdateCells( HDC hdc )
{
    int x, y;

    int i_inst = wTopRow * wNumCols;
    for(  y= wTopRow; y<wTopRow+wVisibleRows; y++ )
    {
        for(  x=wLeftCol; x<wLeftCol+wVisibleCols; x++ )
        {
            RECT rect;
    
            GetCellRect( MAKECELL( x, y ), rect );
            InflateRect( &rect, -2, -2 );

            DrawCharPoints( hdc, rect, pDrawData->cur_char, i_inst++ );
        }  
    }
}



//
//  OWL's encapsulation of the WM_PAINT Message.
//  Update display according to viewing mode.
//
void TGridWindow::Paint( HDC PaintDC, PAINTSTRUCT _FAR & )
{
    if ( fInFocus )
    {
        RECT rect;
        HPEN hpen = CreatePen( PS_INSIDEFRAME,
                               2*GetSystemMetrics(SM_CXBORDER),
                               GetSysColor( COLOR_ACTIVEBORDER ));
        HPEN hpenOld = SelectPen( PaintDC, hpen );
        GetWindowRect( HWindow, &rect );
        Rectangle( PaintDC, 0, 0, rect.right-rect.left,
                   rect.bottom-rect.top );
        SelectPen( PaintDC, hpenOld );
        DeletePen( hpen );
    }

    switch( vMode )
    {
        case gridMode:
                DrawGrid( PaintDC );
                if ( IsCellVisible( wSelectedCell ))
                     FrameCell( PaintDC, wSelectedCell );

                if (((PTMainWindow)Parent)->IsData() )
                {
                    UpdateCells( PaintDC );
                }
                break;

        case blownUpMode:
                if (((PTMainWindow)Parent)->IsData() )
                {
                    RECT rect;
                     
                    GetClientRect( HWindow, &rect );
                    InflateRect( &rect, -2, -2 );
                    DrawCharLines( PaintDC,
                                   rect,
                                   pDrawData->cur_char,
                                   pDrawData->cur_inst,
                                   pDrawData->nScaleFactor );
                }
                break;

        default:
                MessageBeep( 0 );
                 break;   
    }
}



//
//  Set Number of Rows....
//
void TGridWindow::SetNumRows( WORD nRow )
{
    wNumRows = nRow; 
}



//
//  Set Number of Columns
//
void TGridWindow::SetNumCols( WORD nCol )
{
    wNumCols = nCol;
}



//
//  Returns ID of Cell that contains POINT pt
//
WORD TGridWindow::CellFromPoint( POINT pt )
{
    return MAKECELL( 
                   (( pt.x - wBdrWidth )/wCellWidth)  + wLeftCol, 
                   (( pt.y - wBdrHeight)/wCellHeight) + wTopRow
                   );
}



//
//  Returns bounding rectangle of specified Cell
//
void TGridWindow::GetCellRect( WORD wCell, RECT & rect )
{
    rect.left   = ( CELLCOL(wCell) - wLeftCol) * wCellWidth + wBdrWidth;
    rect.top    = ( CELLROW(wCell) - wTopRow ) * wCellHeight+ wBdrHeight;
    rect.right  = rect.left + wCellWidth;
    rect.bottom = rect.top  + wCellHeight;
}



//
//  Returns ID of Cell right of currently selected one.
//
WORD TGridWindow::CellRight( void )
{
    return((CELLCOL( wSelectedCell ) < wNumCols-1) ?
            MAKECELL(CELLCOL(wSelectedCell)+1, CELLROW(wSelectedCell)) : -1);
}



//
//  Returns ID of Cell left of currently selected one.
//
WORD TGridWindow::CellLeft( void )
{
    return( CELLCOL( wSelectedCell ) ?
            MAKECELL(CELLCOL(wSelectedCell)-1, CELLROW(wSelectedCell)) : -1);
}



//
//  Returns ID of Cell above currently selected one.
//
WORD TGridWindow::CellUp( void )
{
    return( CELLROW( wSelectedCell ) ?
            MAKECELL(CELLCOL(wSelectedCell), CELLROW(wSelectedCell)-1) : -1);

}



//
//  Returns ID of Cell below currently selected one.
//
WORD TGridWindow::CellDown( void )
{
    return((CELLROW( wSelectedCell ) < wNumRows-1 ) ?
            MAKECELL(CELLCOL(wSelectedCell), CELLROW(wSelectedCell)+1) : -1);
}




//
//  Scrolls Grid Up
//
void TGridWindow::ScrollUp()
{
    if ( wTopRow )
        wTopRow--;
    UpdateView();
}



//
//  Scrolls Grid Down
//
void TGridWindow::ScrollDown()
{
    if ( wTopRow < wNumRows-wVisibleRows )
        wTopRow++;
    UpdateView();
}



//
//  Scrolls Grid Right
//
void TGridWindow::ScrollRight()
{
    if ( wLeftCol )
        wLeftCol--;
    UpdateView();
}



//
//  Scrolls Grid Left
//
void TGridWindow::ScrollLeft()
{
    if ( wLeftCol < wNumCols-wVisibleCols )
        wLeftCol++;
    UpdateView();
}




//
//  Function to indicate whether a Cell is currently visible
//
BOOL TGridWindow::IsCellVisible( WORD wCell )
{
    return( CELLCOL(wCell) >= wLeftCol  &&  CELLROW(wCell) >= wTopRow  &&
            CELLCOL(wCell) <  wVisibleCols+wLeftCol  &&
            CELLROW(wCell) <  wVisibleRows+wTopRow      );
}



//
//  Selects a New Cell from the Grid
//
void TGridWindow::SelectNewCell( WORD wCell )
{
    if ( wCell != wSelectedCell  &&  wCell != WORD(-1)  &&
         CELLROW( wCell ) < wNumRows  &&  CELLCOL( wCell ) < wNumRows )
    {
        ((PTMainWindow)Parent)->SetCurChar( pDrawData->cur_char,
                             CELLROW( wCell )*wNumCols +
                             CELLCOL( wCell )
                                          );
        if ( !IsCellVisible( wCell ))
        {
            if      ( CELLCOL( wCell ) < wLeftCol )
            {
                ScrollRight();
            }
            else if ( CELLCOL( wCell ) >= wVisibleCols+wLeftCol )
            {
                ScrollLeft();
            }
            else if ( CELLROW( wCell ) < wTopRow  )
            {
                ScrollUp();
            }
            else
            {
                ScrollDown();
            }
        }
        else
        {
            RECT rect;

            GetCellRect( wSelectedCell, rect );
            InvalidateRect( HWindow, &rect, TRUE );

            GetCellRect( wCell, rect );
            InvalidateRect( HWindow, &rect, TRUE );
        }
        wSelectedCell = wCell;
    }
}



//
//  Draws the Grid of Cells
//
void TGridWindow::DrawGrid( HDC hdc )
{
    RECT rc;
    WORD wGridWidth = wCellWidth *wVisibleCols;
    WORD wGridHeight= wCellHeight*wVisibleRows;

    for( int i=0; i<=wVisibleRows; i++ )    // Draw the H. Lines
    {
        MoveTo( hdc, wBdrWidth,  i*wCellHeight + wBdrHeight );
        LineTo( hdc, wGridWidth+wBdrWidth, i*wCellHeight + wBdrHeight );
    }
    for( i=0; i<=wVisibleCols; i++ )       // Draw the V. Lines
    {
        MoveTo( hdc, i*wCellWidth + wBdrWidth, wBdrHeight );
        LineTo( hdc, i*wCellWidth + wBdrWidth, wGridHeight + wBdrHeight );
    }
}




//
//  Draws a frame to visually indicate the currently
//  Selected Cell.
//
void TGridWindow::FrameCell( HDC hdc, WORD wCell )
{
    RECT rc;

    GetCellRect( wCell, rc );

    rc.top++;  rc.left++;
    InflateRect( &rc, -1, -1 );
    FrameRect( hdc, &rc, GetStockBrush(DKGRAY_BRUSH) );
}




//
//  Get Currently Selected Instance - Used only in Alphabet Mode
//
int TGridWindow::GetCurInst()
{
    return( 0 );
}



//
//  Request Arrow keys and WM_CHAR Messages.
//
void TGridWindow::WMGetDlgCode( RTMessage msg )
{
    msg.Result = DLGC_WANTARROWS | DLGC_WANTCHARS;
}



//
//  Handles Background's appearance.
//
void TGridWindow::WMEraseBkgnd( RTMessage msg )
{
    HDC hdc = (HDC)msg.WParam;
    RECT rect;
    GetClientRect( HWindow , &rect );

    HBRUSH hb = CreateSolidBrush( pDrawData->rgbBackGroundColor );
    UnrealizeObject( hb );

    FillRect( hdc , &rect , hb );
    DeleteObject( hb );
    msg.Result = TRUE;
}



//
//  Processes selection of new Cell via Mouse LButtonDown.
//
void TGridWindow::WMLButtonDown( RTMessage msg ) 
{
    if ( vMode == gridMode )
    {
        SelectNewCell( CellFromPoint( MAKEPOINT( msg.LParam )));
    }
    if ( !fInFocus )
        PostMessage( Parent->HWindow, WM_NEXTDLGCTL, ( WPARAM )HWindow,
                                                     ( LPARAM )TRUE );

}



//
//  Processes Focus gaining/loosing
//
void TGridWindow::WMSetFocus( RTMessage msg )
{
    if ( msg.WParam )
    {
        fInFocus = TRUE;
        InvalidateRect( HWindow, NULL, FALSE );
    }
    else
    {
        fInFocus = FALSE;
        InvalidateRect( HWindow, NULL, TRUE );        
    }

}



//
//  Switches back to GridMode.
//
void TGridWindow::GridMode()
{
    vMode = gridMode;
    ((PTMainWindow)Parent)->EnableScaleScrollBar( FALSE );
    UpdateView();
}



//
//  Switches to BlownupMode.
//
void TGridWindow::BlownUpMode()
{
    vMode = blownUpMode;
    ((PTMainWindow)Parent)->EnableScaleScrollBar( TRUE );
    UpdateView();
}



//
//  Processes Double Click to enable
//  GridMode <-> BlownUpMode switching.
//
void TGridWindow::WMLButtonDblClk( RTMessage msg )
{
    switch( vMode )
    {
        case gridMode:
                 if ( IsCellVisible( 
                      CellFromPoint( MAKEPOINT( msg.LParam ) )
                                   )  &&
                      (((PTMainWindow)Parent)->IsData())
                    )
                 {
                    BlownUpMode();
                 }
                 break;

        case blownUpMode:
                 GridMode();
                 break;
        default:
                break;
    }
}



//
//  Processes KeyPresses to enable user to navigate thru'
//  Cells using Keyboard...
//
void TGridWindow::WMKeyDown( RTMessage msg ) 
{
    if ( vMode == gridMode )
    {
        switch( msg.WParam )
        {
            case VK_RIGHT:
                      SelectNewCell( CellRight() );
                      break;

            case VK_LEFT:
                      SelectNewCell( CellLeft() );
                      break;

            case VK_UP:
                      SelectNewCell( CellUp() );
                      break;

            case VK_DOWN:
                      SelectNewCell( CellDown());
                      break;

            case VK_RETURN:
                      if (((PTMainWindow)Parent)->IsData())
                          BlownUpMode();
                      break;

            default:
                      break;
        }
    }

    if ( vMode == blownUpMode  &&  msg.WParam == VK_RETURN )
    {
        GridMode();
    }
    DefWndProc( msg ); 
}



//
//  Processes WM_CHAR Messages to allow user to alter
//  the Character to be displayed by simply pressing
//  the requested one... 
//
void TGridWindow::WMChar( RTMessage msg )
{
    if ( (int)msg.WParam < MAX_CHARS   &&
         (((PTMainWindow)Parent)->IsData()))
    {
        if ( msg.WParam != pDrawData->cur_char )
        {
            ((PTMainWindow)Parent)->SetCurChar(
                                               (int)msg.WParam,
                                                pDrawData->cur_inst
                                              );
            UpdateView();
        }
    }
}



