// ====  Sample Plug Definitions  ====
//
// Copyright (C) 1992  William G. Wong
//
// Version	1.01	Sept 1992
//		1.00	June 1992

#include "plugs.hpp"


// ====  Class definitions  ====
//
// These definitions are normally found in a header file.
//
// ----  Simple abstract plug that can link to objects link itself  ----

class plugXBase		// simple accessible class
{
  public :
    int plugXBaseCommonItem ;
} ;


// ----  Declare plugs  ----

DECLARE_CONNECTION (plugXA, plugXBase
		   ,plugYA, emptyClass
		   )

DECLARE_PLUG(plugX,plugXA)
DECLARE_PLUG(plugY,plugYA)

DECLARE_REDIRECTOR(plugXRedirector,plugXA)


// ----  Dynamic allocation example using multiPlugX and dynamicPlugX  ----
//
// This pair of plug classes work like plugX except that a single
// plug class, multiPlugX, creates a new object, of dynamicPlugX
// class, each time a connection is made. The dynamicPlugX
// object is freed when the connection is broken.
//
// Both <= and connectedto must be redefined because a connection
// will come through one or the other depending upon which side the
// multiPlugX object is when the initial connection is initiated.
//
// This version of does not keep track of the newly created objects.
// They free themselves when disconnected. A reference to the creating
// object can be maintained in dynamicPlugX objects if the object
// must remove itself from a list kept by multiPlugX.

class multiPlugX ;

class dynamicPlugX : public plugX
{
  protected:
    multiPlugX & myParent ;	// can be used in disconnectedFree
				// to remove itself from a list
  public:
    dynamicPlugX ( multiPlugX & parent ) : myParent(parent) {}

    void connectFailed  () { disconnectFree () ; }
    void disconnectFree () { delete this ; }
};


class multiPlugX : public plugX
{
  public:
    plugXA * getplug() { return new dynamicPlugX (* this) ;}
				// add plug to a list at this point
				// if the parent tracks the children
};


// ----  sample1 to sample2 connection  ----

class sample1Base
{
  public:
    plugX x, y ;
} ;


// ----  Declare plugs  ----
//
// Declares non-abstract plugs (abstract plugs are hidden).
// Redirectors can be defined using plugs name followed by
// an A which is the name of the abstract classes defined
// within the macros.

DECLARE_PLUGS_AND_REDIRECTORS
  ( sample1, sample1Base, sample1Redirector
  , sample2, emptyClass,  sample2Redirector
  )


// ----  A slightly more complex plug base  ----

class myPlugBase
{
  public:
    int i ;
    virtual void read  ( int x ) { i = x ; } ;
    virtual void write ( int x ) { i = x ; } ;

    myPlugBase () { i = 1 ; } ;
} ;


// ----  Declare plugs  ----

DECLARE_PLUGS ( myPlug1, myPlugBase
              , myPlug2, myPlugBase
	      )


// ----  Implementat plugs  ----
//
// myPlug1a and myPlug1b redifine methods from the base class.

class myPlug1a : public myPlug1
{
  public :
    int i1 ;
    virtual void write  ( int x ) { i1 = x ; } ;
} ;


class myPlug2a : public myPlug2
{
  public :
    int i2 ;
    virtual void write  ( int x ) { i2 = x ; } ;
} ;


// ====  Sample program  ====

void test ( myPlug1 & pp1 )
{
  plugX x1, x2, x3 ;

  plugXA * ppx ;

  plugXRedirector rx1, rx2 ;

  plugY y1, y2, y3 ;

  sample1 s1 ;
  sample2 s2 ;

  myPlug1a p1 ;
  myPlug2a p2 ;

  multiPlugX mpx ;


  // ----  General connections  ----

  x1 <= y1 ;			// connect x1 and y1
  y1 <= x2 ;			// disconnect x1
				// connect    y1 and x2
  s1 <= s2 ;			// connect types must match


  // ----  Compound linkages  ----

  s2 -> x <= y1 ;		// linking s1.x to y1
  s2 -> y <= y2 ;		// linking s2.y to y2


  // ----  Redirector examples  ----

  rx1 = & x1 ;			// setup redirector
  rx1 <= y1 ;			// connect x1 to y1 through r1

  y1 -> plugXBaseCommonItem = 1 ;
				// access item at other end
  rx1.addBefore ( & y2, & x3 ) ;// rx1 redirects x3, x1 connected to y2
				//  and x3 is attached to y1

  rx2 = & x3 ;			// setup a different redirector type
  rx2 <= y2 ;			// connects y2 to x3 through r2

  rx2 = 0 ;			// reset redirector


  // ----  Multiplug examples  ----
  //
  // This generates two new objects and deletes them when they are
  // no longer of use. Note how pp is used to keep around a reference
  // to (mp<=a) (the one created for a) which is later used to link to c.

  mpx <= y1 ;			// y1 connected to new object
  mpx <= y2 ;			// y2 connected to different object

  ppx = - y1 ;			// disconnect but keep around object

  (* ppx) <= y3 ;		// link object to y3

  -- x2 ;		      	// free up the other end of x2

  ppx = - y3 ;			// - disconnects and returns pointer
  -- * ppx ;			// -- * frees result


  // ----  Plugs with different variables or methods  ----

  p1  <= p2 ;

  p1 -> read ( p1 -> i ) ;	// using read and i from p2

  pp1 <= p2 ;			// connects pp1 to p2
				// disconnects p1


  // ---- Explicit disconnects  ----

  -- s2 ;			// forced disconnect
  -- s1 ;

  if (x1.connected())		// determine if connection exists
    -- x1 ;


  // ---- Implicit disconnects for locally defined objects  ----
}


void main ()
{
  myPlug1a p1 ;

  test ( p1 ) ;
}


// ====  End of PLUGS.CPP  ====


