// ====  PLUGS.HPP    ====
//
// Copyright (C) 1992  William G. Wong
//
// V1.01 Sept 1992	getplug() and connectFailed() added to improve
//			  dynamic allocation definition and procedures
//			Added beforeDisconnect in <= if disconnecting (bug)
// V1.00 June 1992	Original version

#if !defined(_PLUGS_HPP)
#define _PLUGS_HPP

// ----  Description of plug interface  ----
//
// Defining a plug class is not covered in these comments.
// A plug class uses the following operations:
//
//	A <= B			connects A to B, does -- A and -- B first
//	-  A			disconnects A, but will not free otherEnd
//				  returns otherEnd
//	-- A             	disconnects A, will free otherEnd
//	A.connected()       	returns 0 if A is disconnected
//	A.usable()		returns 0 if not usable
//				  (i.e. uninitialized redirector)
//
//	A -> V			access a member of the object at
//				the other end of A, normally assumes there
//				is something at the other end
//
//	R = & A			change redirector R's reference to A
//				  returns prior reference
//
//	A.addBefore(B,C)	connects A to B and C to what A was connected to
//				  if A has a redirector then it points to C
//	A.addAfter (B,C)	like addBefore
//				  if A has a redirector it does not change
//
// ----  Plug Class Definition Macros  ----
//
// DECLARE_PLUGS (plug1,base1,plug2,base2)
//
// DECLARE_PLUGS_AND_REDIRECTORS (plug1,base1,redir1,plug2,base2,redir2)
//
//
// ----  Specialized Plug Class Definition Macros  ----
//
// DECLARE_CONNECTION (abstractPlug1,base1,abstractPlug2,base2)
//
// DECLARE_PLUG (plug,abstractPlug)
//
// DECLARE_REDIRECTOR (redirector,abstractPlug)
//
//
// ----  plugCheck alternate definition  ----
//
// Define a check macro if you want to check for invalid access
// using ->.

#if !defined(plugCheck)
#define plugCheck(ptr)
#endif


// ----  Simple plug macro definitions  ----

#if !defined (emptyClass)
class emptyClass {} ;		// for simple end plugs
#endif

#if !defined(BOOL)
typedef int BOOL ;
#endif
#if !defined(FALSE)
#define FALSE 0
#endif
#if !defined(TRUE)
#define TRUE 1
#endif

#define DECLARE_PLUGS_AND_REDIRECTORS(p1,b1,r1,p2,b2,r2)\
DECLARE_PLUGS(p1,b1,p2,b2)\
DECLARE_REDIRECTOR(r1,p1##A)\
DECLARE_REDIRECTOR(r2,p2##A)

#define DECLARE_PLUGS(p1,b1,p2,b2)\
DECLARE_CONNECTION(p1##A,b1,p2##A,b2)\
DECLARE_PLUG(p1,p1##A)\
DECLARE_PLUG(p2,p2##A)

#define DECLARE_CONNECTION(p1,b1,p2,b2)\
class p1##HiddenRedirector;\
class p2##HiddenRedirector;\
class p2;\
DECLARE_ABSTRACT_PLUG(p1,b1,p1##HiddenRedirector,p2)\
DECLARE_ABSTRACT_PLUG(p2,b2,p2##HiddenRedirector,p1)\
DECLARE_ABSTRACT_REDIRECTOR(p1##HiddenRedirector,p1)\
DECLARE_ABSTRACT_REDIRECTOR(p2##HiddenRedirector,p2)

#define DECLARE_ABSTRACT_PLUG(p1,b1,r1,p2)\
class p1 : public b1\
{\
 public:\
  typedef r1 * pRedirector ;\
  typedef p2 * pOtherPlug ;\
  typedef p2 & rOtherPlug ;\
  friend class r1;\
  friend class p2;\
  virtual ~p1 () {}\
  virtual BOOL isRedirector     () = 0 ;\
  virtual BOOL connected        () = 0 ;\
  virtual BOOL usable           () = 0 ;\
  virtual void operator <=      (rOtherPlug newEnd) = 0 ;\
  virtual pOtherPlug operator ->() = 0 ;\
  virtual pOtherPlug operator - () = 0 ;\
  virtual void operator --      () = 0 ;\
  virtual void addBefore        ( pOtherPlug inner, p1 * newEnd ) = 0 ;\
  virtual void addAfter         ( pOtherPlug inner, p1 * newEnd ) = 0 ;\
\
  virtual void afterConnect     () {}\
  virtual void beforeDisconnect () {}\
  virtual pRedirector getRedirector() { return 0 ; }\
  virtual void setRedirector    ( pRedirector newRedirector ) {}\
  virtual void connectFailed    () {}\
  virtual p1 * getplug          () = 0 ;\
  virtual void connectto        ( pOtherPlug newEnd ) = 0 ;\
  virtual void disconnectedFree () {}\
  virtual void disconnected     () {}\
};

#define DECLARE_ABSTRACT_REDIRECTOR(r1,p1)\
class r1 : public p1\
{\
 public:\
  virtual p1 * operator = ( p1 * newPlug ) = 0 ;\
  virtual void resetRedirection () = 0 ;\
};


#define DECLARE_PLUG(c1,p1)\
class c1 : public p1\
{\
 public:\
  typedef p1 * pPlug;\
  typedef p1::pRedirector pRedirector;\
  typedef p1::pOtherPlug  pOtherPlug;\
  typedef p1::rOtherPlug  rOtherPlug;\
 protected:\
  pRedirector myRedirector ;\
  pOtherPlug  myPlug ;\
 public:\
	   c1 () {myPlug=0;myRedirector=0;}\
  virtual ~c1 ()\
    {\
      pOtherPlug prior = - (* this) ;\
      if (prior) -- (* prior) ;\
    }\
  virtual BOOL  isRedirector     () {return TRUE;}\
  virtual BOOL  connected        () {return myPlug != 0 ;}\
  virtual BOOL  usable           () {return TRUE;}\
  virtual void  operator <=      (rOtherPlug newEnd)\
    {\
      pOtherPlug prior = - (* this) ;\
      if (prior) -- (* prior) ;\
      if (& newEnd)\
      {\
        pPlug newPlug = getplug () ;\
        if (newPlug)\
        {\
          pOtherPlug otherPlug = newEnd.getplug ();\
          if (otherPlug)\
          {\
            newPlug -> connectto (otherPlug);\
	    otherPlug -> connectto (newPlug);\
	    (*newPlug).afterConnect();\
	    (*otherPlug).afterConnect();\
          }\
          else\
	    newPlug->connectFailed();\
        }\
      }\
      else\
        connectFailed ();\
    }\
  virtual pOtherPlug operator -> () {return myPlug;}\
  virtual pOtherPlug operator -  ()\
    {\
      pOtherPlug prior = myPlug ;\
      if (prior)\
      {\
        beforeDisconnect();\
        prior -> beforeDisconnect();\
        disconnected () ;\
        prior -> disconnected () ;\
      }\
      return prior ;\
    }\
  virtual void  operator --      () \
    {\
      pOtherPlug prior = - (* this) ;\
      if (prior) -- (* prior) ;\
      disconnectedFree();\
    }\
  virtual void  addBefore        ( pOtherPlug inner, p1 * newEnd )\
    {\
      pRedirector oldRedirector = myRedirector ;\
      addAfter ( inner, newEnd ) ;\
      if (oldRedirector!=0)\
      {\
        oldRedirector -> resetRedirection () ;\
        (* oldRedirector) = newEnd ;\
      }\
    }\
  virtual void  addAfter         ( pOtherPlug inner, p1 * newEnd )\
    {\
      if (myPlug && newEnd)\
        (* newEnd) <= (* (- (* this))) ;\
      (* this) <= (* inner) ;\
    }\
  virtual pRedirector getRedirector() {return myRedirector;}\
  virtual void  setRedirector    ( pRedirector newRedirector )\
    {myRedirector=newRedirector;}\
  virtual pPlug getplug          () { return this ; }\
  virtual void  connectto        ( pOtherPlug newEnd )\
    {\
      if (myPlug)\
        -- (* myPlug) ;\
      myPlug = newEnd ;\
    }\
  virtual void  disconnectedFree () {myPlug=0;}\
  virtual void  disconnected     () {myPlug=0;}\
};


#define DECLARE_REDIRECTOR(r1,p1)\
class r1 : public p1##HiddenRedirector\
{\
 public:\
  typedef p1 * pPlug;\
  typedef p1::pRedirector pRedirector;\
  typedef p1::pOtherPlug  pOtherPlug;\
  typedef p1::rOtherPlug  rOtherPlug;\
 protected:\
  pRedirector myRedirector ;\
  pPlug       myPlug ;\
 public:\
  virtual pPlug operator =       ( pPlug newPlug )\
    {\
      r1::pPlug prior = myPlug ;\
      resetRedirection () ;\
      if (newPlug != 0)\
      {\
        myPlug = newPlug ;\
        newPlug -> setRedirector((pRedirector)this) ;\
      }\
      return prior ;\
    }\
  virtual void resetRedirection  () {myPlug=0;}\
	   r1                    (pPlug newPlug=0)\
	    {myPlug=newPlug;myRedirector=0;}\
  virtual ~r1                    () {(* this) = 0;}\
  virtual BOOL  isRedirector     () {return TRUE;}\
  virtual BOOL  connected        ()\
    { return (myPlug == 0) ? FALSE : myPlug -> connected () ; }\
  virtual BOOL  usable           ()\
    { return (myPlug == 0) ? FALSE : myPlug -> usable () ; }\
  virtual void  operator <=      (rOtherPlug newEnd)\
    {if (myPlug) (* myPlug) <= newEnd ;}\
  virtual pOtherPlug operator -> ()\
    {return (myPlug) ? myPlug -> operator -> () : 0 ;}\
  virtual pOtherPlug operator -  ()\
    {return (myPlug) ? - * myPlug : 0 ;}\
  virtual void  operator --      () {if (myPlug) -- * myPlug ;}\
  virtual void  addBefore        ( pOtherPlug inner, pPlug newEnd )\
    {\
      if (myPlug)\
      {\
        myPlug -> setRedirector (0) ;\
        (* myPlug).addBefore ( inner, newEnd ) ;\
      }\
      if (newEnd)\
        newEnd -> setRedirector ((pRedirector) this ) ;\
       myPlug = newEnd ;\
    }\
  virtual void  addAfter         ( pOtherPlug inner, pPlug newEnd )\
    { if (myPlug) (* myPlug).addAfter ( inner, newEnd ) ; }\
  virtual pRedirector getRedirector() {return myRedirector;}\
  virtual void  setRedirector    ( pRedirector newRedirector )\
    {myRedirector=newRedirector;}\
  virtual pPlug getplug          ()\
    {return (myPlug) ? myPlug->getplug() : 0 ;}\
  virtual void  connectto        ( pOtherPlug newEnd )\
    {if (myPlug) myPlug->connectto(newEnd);}\
  virtual void  connectFailed    ()\
    {if (myPlug) myPlug->connectFailed();}\
  virtual void  disconnectedFree ()\
    {if (myPlug) myPlug->disconnectedFree();}\
  virtual void  disconnected     ()\
    {if (myPlug) myPlug->disconnected();}\
  virtual void  afterConnect     ()\
    {if (myPlug) myPlug->afterConnect();}\
  virtual void  beforeDisconnect ()\
    {if (myPlug) myPlug->beforeDisconnect();}\
};


#endif

// ====  End of PLUGS.HPP  ====


